import os
import sys
import subprocess
import xml.dom.minidom
from xml.dom.minidom import parseString
import xml.etree.ElementTree as ET
import re
import json
import tempfile
import signal
import time
from io import BytesIO
import tarfile
import getpass
import base64
import threading
import logging
from functools import lru_cache
from urllib.parse import urlencode
import settings
import xml.parsers as parsers
import xmltodict
import abc
from pcs.common.reports import ReportProcessor
from pcs.common.reports.item import ReportItem
#from pcs.common.reports import codes as report_codes
from pcs.lib.external import CommandRunner
#import reports
import pam
from errors import ReportItemSeverity
from pcs.cli.reports import (
    output as reports_output,
    process_library_reports,
    ReportProcessorToConsole,
)


usefile = False
filename = ""
pcs_options = {}

_find_unsafe = re.compile(r'[^\w@%+=:,./-]', re.ASCII).search


def get_cluster_property_from_xml(etree_el):
    """
    Commandline options: no options
    """
    prop = {
        "name": etree_el.get("name", ""),
        "shortdesc": "",
        "longdesc": "",
    }
    for item in ["shortdesc", "longdesc"]:
        item_el = etree_el.find(item)
        if item_el is not None and item_el.text is not None:
            prop[item] = item_el.text

    content = etree_el.find("content")
    if content is None:
        prop["type"] = ""
        prop["default"] = ""
    else:
        prop["type"] = content.get("type", "")
        prop["default"] = content.get("default", "")

    if prop["type"] == "enum":
        prop["enum"] = []
        if prop["longdesc"]:
            values = prop["longdesc"].split("  Allowed values: ")
            if len(values) == 2:
                prop["enum"] = values[1].split(", ")
                prop["longdesc"] = values[0]
        if prop["default"] not in prop["enum"]:
            prop["enum"].append(prop["default"])

    if prop["longdesc"] == prop["shortdesc"]:
        prop["longdesc"] = ""
    return prop


def shell_quote(s):
    """Return a shell-escaped version of the string *s*."""
    if not s:
        return "''"
    if _find_unsafe(s) is None:
        return s

    # use single quotes, and put single quotes into double quotes
    # the string $'b is then quoted as '$'"'"'b'
    return "'" + s.replace("'", "'\"'\"'") + "'"


class SimpleReportProcessorInterface(abc.ABC):
    def report(self, report_item):
        return self.report_list([report_item])

    @abc.abstractmethod
    def report_list(self, report_list):
        raise NotImplementedError()


def get_report_processor():
    return ReportProcessorToConsole(debug=("--debug" in pcs_options))


# Run command, with environment and return (output, retval)
# DEPRECATED, please use lib.external.CommandRunner via utils.cmd_runner()
def run(args,
        ignore_stderr=False,
        string_for_stdin=None,
        env_extend=None,
        binary_output=False):
    """
    Commandline options:
      * -f - CIB file (effective only for some pacemaker tools)
      * --debug
    """
    print(env_extend)
    print(args)

    if not env_extend:
        env_extend = dict()
    env_var = env_extend
    env_var.update(dict(os.environ))
    env_var["LC_ALL"] = "C"
    if usefile:
        env_var["CIB_file"] = filename
        touch_cib_file(filename)

    command = args[0]
    if (command[0:3] == "crm"
            or command in ["cibadmin", "iso8601", "stonith_admin"]):
        args[0] = settings.pacemaker_binaries + command
    elif command[0:8] == "corosync":
        args[0] = settings.corosync_binaries + command

    try:
        if "--debug" in pcs_options:
            print("Running: " + " ".join(args))
            if string_for_stdin:
                print("--Debug Input Start--\n" + string_for_stdin)
                print("--Debug Input End--")

        # Some commands react differently if you give them anything via stdin
        if string_for_stdin is not None:
            stdin_pipe = subprocess.PIPE
        else:
            stdin_pipe = subprocess.DEVNULL

        # pylint: disable=subprocess-popen-preexec-fn
        p = subprocess.Popen(
            args,
            stdin=stdin_pipe,
            stdout=subprocess.PIPE,
            stderr=(subprocess.PIPE if ignore_stderr else subprocess.STDOUT),
            preexec_fn=subprocess_setup,
            close_fds=True,
            env=env_var,
            # decodes newlines and in python3 also converts bytes to str
            universal_newlines=(not binary_output))
        output, dummy_stderror = p.communicate(string_for_stdin)
        returnVal = p.returncode
        if "--debug" in pcs_options:
            print("Return Value: {0}".format(returnVal))
            print(("--Debug Output Start--\n{0}".format(output)).rstrip())
            print("--Debug Output End--")
            print()
    except OSError as e:
        print(e.strerror)
        err("unable to locate command: " + args[0])

    return output, returnVal


def get_cib(scope=None):
    """
    Commandline options:
      * -f - CIB file
    """
    command = ["cibadmin", "-l", "-Q"]
    if scope:
        command.append("--scope=%s" % scope)
    output, retval = run(command)
    if retval != 0:
        if retval == 105 and scope:
            err("unable to get cib, scope '%s' not present in cib" % scope)
        else:
            err("unable to get cib")
    return output


@lru_cache()
def cmd_runner():
    """
    Commandline options:
      * -f - CIB file
    """
    env_vars = dict()
    if usefile:
        env_vars["CIB_file"] = filename
    env_vars.update(os.environ)
    env_vars["LC_ALL"] = "C"
    return CommandRunner(logging.getLogger("pcs"), get_report_processor(),
                         env_vars)


# 获取集群当前属性
def get_cluster_properties():
    status, stdout = run_cmd("cibadmin --query --scope crm_config")
    json_str = json.dumps(xmltodict.parse(stdout))
    res_json = json.loads(json_str)['crm_config']
    nvpairs = res_json['cluster_property_set']
    if isinstance(nvpairs, list):
        cluster_properties = {}
        for nvpair in nvpairs:
            if "nvpair" in nvpair:
                if isinstance(nvpair["nvpair"], list):
                    for nvpa in nvpair["nvpair"]:
                        cluster_properties[nvpa["@name"]] = nvpa["@value"]
                else:
                    cluster_properties[nvpair["nvpair"]["@name"]] = nvpair["nvpair"]["@value"]
    else:
        nvpairs = res_json['cluster_property_set']['nvpair']
        cluster_properties = {}
        for nvpair in nvpairs:
            cluster_properties[nvpair["@name"]] = nvpair["@value"]
    return cluster_properties


def get_cluster_properties_definition():
    """
    Commandline options: no options
    """

    cluster_properties = get_cluster_properties()

    enable_list = [
        "node-health-green", "stonith-enabled", "symmetric-cluster",
        "maintenance-mode", "node-health-yellow", "node-health-yellow",
        "no-quorum-policy", "node-health-red", "node-health-strategy",
        "default-resource-stickiness", "start-failure-is-fatal"
    ]

    sources = [{
        "name": "pacemaker-schedulerd",
        "path": settings.pacemaker_schedulerd,
    }, {
        "name": "pacemaker-controld",
        "path": settings.pacemaker_controld,
    }, {
        "name": "pacemaker-based",
        "path": settings.pacemaker_based,
    }]
    data = {}
    for source in sources:
        stdout, stderr, retval = cmd_runner().run([source["path"], "metadata"])
        if retval != 0:
            err("unable to run {0}\n{1}".format(source["name"], stderr))
        try:
            etree = ET.fromstring(stdout)
            for e in etree.findall("./parameters/parameter"):
                prop = get_cluster_property_from_xml(e)
                print(prop)
                name = prop['name']
                if name in enable_list:
                    prop['content'] = {}
                    if name in cluster_properties:
                        prop['value'] = cluster_properties[name]
                    else:
                        if name == "node-health-green":
                            prop['value'] = 0
                        if name == "stonith-enabled":
                            prop['value'] = "true"
                        if name == "symmetric-cluster":
                            prop['value'] = "true"
                        if name == "maintenance-mode":
                            prop['value'] = "false"
                        if name == "node-health-yellow":
                            prop['value'] = "0"
                        if name == "node-health-red":
                            prop['value'] = "0"
                        if name == "no-quorum-policy":
                            prop['value'] = "ignore"
                        if name == "node-health-strategy":
                            prop['value'] = "none"
                        if name == "start-failure-is-fatal":
                            prop['value'] = "true"
                        if name == "default-resource-stickiness":
                            prop['value'] = 0
                    if prop['longdesc'] != "":
                        prop['longdesc'] = _(prop['longdesc'])
                    prop['content']['default'] = prop['default']
                    prop['content']['type'] = prop['type']
                    if prop['type'] == 'enum':
                        prop['content']['values'] = prop['enum']
                    prop['enabled'] = 1
                    data[name] = prop

        except ET.ParseError as e:
            err("unable to parse {0} metadata definition: {1}".format(
                source["name"], e))
    return data


def run_cmd(cmd):
    try:
        status, output = subprocess.getstatusoutput(cmd)
    except Exception as e:
        status = -1
        output = str(e)
    return status, output


def check_auth(username, password):
    if pam.authenticate(username, password):
        S_PASS = password
        return True
    else:
        return False


def get_resource_info_from_xml(cl, et):
    #从xml格式中获取单个资源信息
    if cl == "group":
        prop = []
        for e in et.findall("primitive"):
            prop.append(e.get("id", ""))
    if cl == "primitive":
        prop = {
            "id": et.get("id", ""),
            "category": et.get("category", ""),
            "class": et.get("class", ""),
            "type": et.get("type", ""),
            "provider": et.get("provider", "")
        }
    if cl == "clone":
        # 如果是组资源
        if et.find("group"):
            return get_resource_info_from_xml("group", et.find("group"))
        else:
            et = et.find("primitive")
            prop = et.get("id", "")
    if cl == "meta" or cl == "inst":
        prop = {}
        for item in et.findall("./nvpair"):
            name = item.get("name")
            value = str(item.get("value"))
            if value == "True":
                value = "true"
            if value == "False":
                value = "false"
            prop[name] = value
    if cl == "operations":
        prop = []
        for item in et.findall("./op"):
            res = {}
            for key, value in item.items():
                res[str(key)] = str(value)
            if item.find("instance_attributes"):
                op_ocf_check_level = get_resource_info_from_xml("inst", item.find("instance_attributes"))
                res.update(op_ocf_check_level)
            prop.append(res)
    return prop


def get_resource_info(ct, xml_data):
    etree = ET.fromstring(xml_data)
    try:
        data = {}
        # 根据分类
        if ct == "primitive":
            data = get_resource_info_from_xml("primitive", etree)
        if ct == "group":
            data["rscs"] = get_resource_info_from_xml("group", etree)
        if ct == "clone":
            data["rsc_id"] = get_resource_info_from_xml("clone", etree)
        # 获取meta_attributes
        e = etree.find("meta_attributes")
        if e:
            prop = get_resource_info_from_xml("meta", e)
            data["meta_attributes"] = prop
        # 获取instance_attributes
        e = etree.find("instance_attributes")
        if e:
            prop = get_resource_info_from_xml("inst", e)
            data["instance_attributes"] = prop

        # 获取actions
        e = etree.find("operations")
        if e:
            prop = get_resource_info_from_xml("operations", e)
            data["actions"] = prop

    except ET.ParseError as e:
        err("Unable to parse resource information: " + e)
    return data


if __name__ == '__main__':
    print(get_cluster_properties_definition())
